import importlib
import logging
import os

from markdown import Extension
from markdown.preprocessors import Preprocessor

log = logging.getLogger(__name__)


class CompiledScript:
    def __init__(self, contents, name):
        self.contents = contents
        self.name = name
        self._code_object = None

    @property
    def compiled(self):
        if self._code_object is None:
            self._code_object = compile(self.contents, filename=self.name, mode='exec')

        return self._code_object


class LoadedScript:
    def __init__(self, path):
        self.path = path
        self.name = os.path.basename(self.path).rsplit('.', 1)[0]
        self._patcher = None

    @property
    def patch(self):
        if self._patcher is None:
            spec = importlib.util.spec_from_file_location(self.name, self.path)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)

            self._patcher = getattr(mod, 'patch', None)
            if not callable(self._patcher):
                log.error('Script `%s` has no callable named `patch`', self.path)
                self._patcher = lambda lines: None

        return self._patcher


class PatcherPreprocessor(Preprocessor):
    def __init__(self, config, md):
        self.raw_script = config.get('script', '')
        if self.raw_script:
            self.raw_script = CompiledScript(self.raw_script, 'script')

        self.scripts = []
        self.location = config.get('location', '')
        if self.location:
            self.location = os.path.normpath(self.location)

            if os.path.isfile(self.location):
                self.scripts.append(LoadedScript(self.location))
            elif os.path.isdir(self.location):
                for entry in sorted(os.listdir(self.location)):
                    path = os.path.join(self.location, entry)
                    if os.path.isfile(path):
                        self.scripts.append(LoadedScript(path))
            else:
                log.error('Unable to load script(s) from %s', self.location)

        super(PatcherPreprocessor, self).__init__()

    def run_raw_script(self, lines):
        local_vars = {'lines': lines}
        exec(self.raw_script.compiled, globals(), local_vars)

        return local_vars['lines']

    def run_scripts(self, lines):
        for script in self.scripts:
            new_lines = script.patch(lines)
            if new_lines:
                lines = new_lines

        return lines

    def run(self, lines):
        new_lines = list(lines)

        if self.raw_script:
            new_lines = self.run_raw_script(new_lines)
        elif self.scripts:
            new_lines = self.run_scripts(new_lines)

        return new_lines


class PatcherExtension(Extension):
    def __init__(self, *args, **kwargs):
        self.config = {
            'script': ['', 'Raw code to execute - Default: ""'],
            'location': ['', 'Base path for script paths or path to a single script - Default: ""'],
        }

        super(PatcherExtension, self).__init__(*args, **kwargs)

    def extendMarkdown(self, md):
        self.md = md
        md.registerExtension(self)
        config = self.getConfigs()
        patcher = PatcherPreprocessor(config, md)
        md.preprocessors.register(patcher, 'mkpatcher', 9000)


def makeExtension(*args, **kwargs):
    return PatcherExtension(*args, **kwargs)



from setuptools import find_packages, setup


def get_version():
    with open('mkpatcher/__meta__.py', 'r') as f:
        for line in f:
            if line.startswith('__version__'):
                return line.strip().split('=')[1].strip(' \'"')


def get_long_description():
    with open('README.md', 'r', encoding='utf-8') as f:
        return f.read()


setup(
    name='mkpatcher',
    version=get_version(),
    description='Python-Markdown extension allowing arbitrary scripts to modify MkDocs input files',
    long_description=get_long_description(),
    long_description_content_type='text/markdown',
    keywords=['markdown', 'extensions', 'mkdocs', 'plugins'],
    url='https://github.com/ofek/mkpatcher',
    project_urls={
        'Documentation': 'https://github.com/ofek/mkpatcher#readme',
        'Source Code': 'https://github.com/ofek/mkpatcher',
        'Bug Tracker': 'https://github.com/ofek/mkpatcher/issues',
    },
    author='Ofek Lev',
    author_email='ofekmeister@gmail.com',
    license='Apache-2.0 OR MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'License :: OSI Approved :: Apache Software License',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: Markup :: HTML',
    ],
    packages=find_packages(include=['mkpatcher', 'mkpatcher.*']),
    python_requires='>=3.6',
    install_requires=['Markdown>=3.2.1'],
)
